"""Kalabash sievefilters handlers."""

from django.urls import reverse
from django.dispatch import receiver
from django.utils.translation import gettext as _

from kalabash.core import signals as core_signals

from . import lib


@receiver(core_signals.extra_user_menu_entries)
def menu(sender, location, user, **kwargs):
    """Return extra menu entry."""
    if location != "options_menu" or not hasattr(user, "mailbox"):
        return []
    return [
        {"name": "sievefilters",
         "label": _("Message filters"),
         "url": reverse("kalabash_sievefilters:index"),
         "img": "fa fa-filter"}
    ]


@receiver(core_signals.user_logout)
def userlogout(sender, request, **kwargs):
    """Close managesieve connection."""
    if not hasattr(request.user, "mailbox"):
        return
    try:
        sc = lib.SieveClient(user=request.user.username,
                             password=request.session["password"])
    except Exception:
        pass
    else:
        sc.logout()



"""AppConfig for stats."""

from django.apps import AppConfig


class SieveFiltersConfig(AppConfig):
    """App configuration."""

    name = "kalabash_sievefilters"
    verbose_name = "Sieve filters editor for Kalabash"

    def ready(self):
        from . import handlers



# coding: utf-8
"""
:mod:`imaputils` --- Extra IMAPv4 utilities
-------------------------------------------
"""

import imaplib
import re
from operator import itemgetter
import socket
import ssl

import six

from django.utils.encoding import smart_bytes
from django.utils.translation import gettext as _

from kalabash.lib import imap_utf7  # NOQA
from kalabash.lib.connections import ConnectionsManager
from kalabash.lib.exceptions import KalabashException, InternalError
from kalabash.parameters import tools as param_tools


class ImapError(KalabashException):

    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return str(self.reason)


@six.add_metaclass(ConnectionsManager)
class IMAPconnector(object):
    """The IMAPv4 connector."""

    list_base_pattern = (
        r'\((?P<flags>.*?)\) "(?P<delimiter>.*)" "?(?P<name>[^"]*)"?'
    )
    list_response_pattern_literal = re.compile(
        r'\((?P<flags>.*?)\) "(?P<delimiter>.*)" \{(?P<namelen>\d+)\}')
    list_response_pattern = re.compile(list_base_pattern)
    listextended_response_pattern = \
        re.compile(list_base_pattern + r'\s*(?P<childinfo>.*)')

    def __init__(self, user=None, password=None):
        self.__hdelimiter = None
        self.conf = dict(
            param_tools.get_global_parameters("kalabash_sievefilters"))
        self.address = self.conf["imap_server"]
        self.port = self.conf["imap_port"]
        self.login(user, password)

    def _cmd(self, name, *args, **kwargs):
        """IMAP command wrapper

        To simplify errors handling, this wrapper calls the
        appropriate method (``uid`` or FIXME) and then check the
        return code. If an error has occured, an ``ImapError``
        exception is raised.

        For specific commands commands (FETCH, ...), the result is
        parsed using the IMAPclient module before being returned.

        :param name: the command's name
        :return: the command's result
        """
        try:
            typ, data = self.m._simple_command(name, *args)
        except imaplib.IMAP4.error as e:
            raise ImapError(e)
        if typ == "NO":
            raise ImapError(data)
        if 'responses' not in kwargs:
            if name not in self.m.untagged_responses:
                return None
            return self.m.untagged_responses.pop(name)
        res = []
        for r in kwargs['responses']:
            if r not in self.m.untagged_responses:
                return None
            res.append(self.m.untagged_responses.pop(r))
        return res

    @property
    def hdelimiter(self):
        """Return the default hierachy delimiter.

        This is a simple way to retrieve the default delimiter (see
        http://www.imapwiki.org/ClientImplementation/MailboxList).

        :return: a string
        """
        if self.__hdelimiter is None:
            data = self._cmd("LIST", '""', '""')
            m = self.list_response_pattern.match(data[0].decode())
            if m is None:
                raise InternalError(
                    _("Failed to retrieve hierarchy delimiter"))
            self.__hdelimiter = m.group('delimiter')
        return self.__hdelimiter

    def refresh(self, user, password):
        """Check if current connection needs a refresh

        Is it really secure?
        """
        if self.m is not None:
            try:
                self._cmd("NOOP")
            except ImapError:
                if hasattr(self, "current_mailbox"):
                    del self.current_mailbox
            else:
                return

        self.login(user, password)

    def login(self, user, passwd):
        """Custom login method

        We connect to the server, issue a LOGIN command. If
        successfull, we try to record a eventuel CAPABILITY untagged
        response. Otherwise, we issue the command.

        :param user: username
        :param passwd: password
        """
        try:
            if self.conf["imap_secured"]:
                self.m = imaplib.IMAP4_SSL(self.address, self.port)
            else:
                self.m = imaplib.IMAP4(self.address, self.port)
        except (socket.error, imaplib.IMAP4.error, ssl.SSLError) as error:
            raise ImapError(_("Connection to IMAP server failed: %s" % error))

        passwd = self.m._quote(passwd)
        data = self._cmd("LOGIN", smart_bytes(user), smart_bytes(passwd))
        self.m.state = "AUTH"
        if "CAPABILITY" in self.m.untagged_responses:
            self.capabilities = (
                self.m.untagged_responses.pop('CAPABILITY')[0]
                .decode().split())
        else:
            data = self._cmd("CAPABILITY")
            self.capabilities = data[0].decode().split()

    def logout(self):
        """Logout from server."""
        try:
            self._cmd("CHECK")
        except ImapError:
            pass
        self._cmd("LOGOUT")
        del self.m
        self.m = None
        if hasattr(self, "current_mailbox"):
            del self.current_mailbox

    def _listmboxes(self, topmailbox, mailboxes):
        """Retrieve mailboxes list."""
        pattern = (
            '"{0}{1}%"'.format(
                topmailbox.encode("imap4-utf-7").decode(), self.hdelimiter)
            if topmailbox else "%"
        )
        resp = self._cmd("LIST", '""', pattern, "RETURN", "(CHILDREN)")
        newmboxes = []
        for mb in resp:
            if not mb:
                continue
            if type(mb) in [list, tuple]:
                flags, delimiter, namelen = (
                    self.list_response_pattern_literal.match(
                        mb[0].decode()).groups()
                )
                name = mb[1][0:int(namelen)]
            else:
                flags, delimiter, name, childinfo = (
                    self.listextended_response_pattern.match(
                        mb.decode()).groups())
            flags = flags.split(" ")
            name = bytearray(name, "utf-8")
            name = name.decode("imap4-utf-7")
            mdm_found = False
            for idx, mdm in enumerate(mailboxes):
                if mdm["name"] == name:
                    mdm_found = True
                    descr = mailboxes[idx]
                    break
            if not mdm_found:
                descr = {"name": name}
                newmboxes += [descr]

            if '\\Marked' in flags or '\\UnMarked' not in flags:
                descr["send_status"] = True
            if '\\HasChildren' in flags:
                if '\\NonExistent' in flags:
                    descr["removed"] = True
                descr["path"] = name
                descr["sub"] = []

        mailboxes += sorted(newmboxes, key=itemgetter("name"))

    def getmboxes(self, user, topmailbox=""):
        """Returns a list of mailboxes for a particular user

        By default, only the first level of mailboxes under
        ``topmailbox`` is returned.

        :param user: a ``User`` instance
        :param topmailbox: the mailbox where to start in the tree
        :return: a list
        """
        if topmailbox:
            md_mailboxes = []
        else:
            md_mailboxes = [
                {"name": "INBOX", "class": "fa fa-inbox"},
                {"name": user.parameters.get_value("drafts_folder"),
                 "class": "fa fa-file"},
                {"name": 'Junk', "class": "fa fa-fire"},
                {"name": user.parameters.get_value("sent_folder"),
                 "class": "fa fa-envelope"},
                {"name": user.parameters.get_value("trash_folder"),
                 "class": "fa fa-trash"}
            ]
        self._listmboxes(topmailbox, md_mailboxes)

        return md_mailboxes


def get_imapconnector(request):
    """Simple shortcut to create a connector

    :param request: a ``Request`` object
    """
    imapc = IMAPconnector(user=request.user.username,
                          password=request.session["password"])
    return imapc



"""Sieve filters constants."""

from django.utils.translation import gettext_lazy as _

HEADER_OPERATORS = [
    ("contains", _("contains"), "string"),
    ("notcontains", _("does not contain"), "string"),
    ("is", _("is"), "string"),
    ("isnot", _("is not"), "string")
]

CONDITION_TEMPLATES = [
    {"name": "Subject",
     "label": _("Subject"),
     "operators": HEADER_OPERATORS},
    {"name": "From",
     "label": _("Sender"),
     "operators": HEADER_OPERATORS},
    {"name": "To",
     "label": _("Recipient"),
     "operators": HEADER_OPERATORS},
    {"name": "Cc",
     "label": _("Cc"),
     "operators": HEADER_OPERATORS},
    {"name": "size", "label": _("Size"),
     "operators": [("over", _("is greater than"), "number"),
                   ("under", _("is less than"), "number")]}
]

ACTION_TEMPLATES = [
    {"name": "fileinto", "label": _("Move message to"),
     "args": [
         {"name": "mailbox", "type": "list",
          "vloader": "userfolders"},
         {"name": "copy", "type": "boolean",
          "label": _("Keep local copy"),
          "value": ":copy"},
     ],
     "args_order": ["copy", "mailbox"]},
    {"name": "redirect", "label": _("Redirect message to"),
     "args": [
         {"name": "address", "type": "string"},
         {"name": "copy", "type": "boolean",
          "label": _("Keep local copy"),
          "value": ":copy"},
     ],
     "args_order": ["copy", "address"]},
    {"name": "reject", "label": _("Reject message"),
     "args": [{"name": "text", "type": "string"}]},
    {"name": "stop", "label": _("Stop processing")},
]



"""Custom template tags."""

from django import template
from django.urls import reverse
from django.template.loader import render_to_string
from django.utils.translation import gettext as _

from kalabash.lib.web_utils import render_actions
from kalabash.lib.templatetags.form_tags import configure_field_classes

from .. import lib

register = template.Library()


@register.simple_tag
def sfilters_menu(user):
    entries = [
        {"name": "newfilterset",
         "img": "fa fa-plus",
         "label": _("New filters set"),
         "url": reverse("kalabash_sievefilters:fs_add"),
         "modal": True,
         "modalcb": "filtersetform_cb"
         }
    ]

    return render_to_string('common/menulist.html',
                            {"entries": entries, "user": user})


@register.simple_tag
def fset_menu(mode, setname):
    entries = []
    if mode == "gui":
        entries += [
            {"name": "newfilter",
             "img": "fa fa-plus",
             "label": _("New filter"),
             "url": reverse("kalabash_sievefilters:filter_add",
                            args=[setname]),
             "modal": True,
             "autowidth": True,
             "modalcb": "filterform_cb"}
        ]
    if mode == "raw":
        entries += [
            {"name": "savefs",
             "img": "fa fa-download-alt",
             "label": _("Save filters set"),
             "url": reverse("kalabash_sievefilters:fs_save",
                            args=[setname])},
        ]

    entries += [
        {"name": "activatefs",
         "img": "fa fa-check",
         "label": _("Activate filters set"),
         "url": reverse("kalabash_sievefilters:fs_activate",
                        args=[setname])},
        {"name": "removefs",
         "img": "fa fa-trash",
         "label": _("Remove filters set"),
         "url": reverse("kalabash_sievefilters:fs_delete", args=[setname])},
        {"name": "downloadfs",
         "img": "fa fa-download",
         "label": _("Download"),
         "url": reverse("kalabash_sievefilters:fs_download", args=[setname])}
    ]

    return render_to_string('common/menulist.html',
                            {"entries": entries})


@register.simple_tag
def filter_actions(setname, f, position, islast):
    actions = [
        {"name": "editfilter",
         "url": reverse(
             "kalabash_sievefilters:filter_change", args=[setname, f["name"]]),
         "img": "fa fa-edit",
         "title": _("Edit filter"),
         "modal": True,
         "autowidth": True,
         "modalcb": "filterform_cb"},
        {"name": "removefilter",
         "url": reverse(
             "kalabash_sievefilters:filter_delete", args=[setname, f["name"]]),
         "img": "fa fa-trash",
         "title": _("Remove this filter")}
    ]
    if position != 1:
        actions += [
            {"name": "movefilter_up",
             "url": reverse(
                 "kalabash_sievefilters:filter_move_up",
                 args=[setname, f["name"]]),
             "img": "fa fa-chevron-up",
             "title": _("Move this filter up")}
        ]
    if not islast:
        actions += [
            {"name": "movefilter_down",
             "url": reverse(
                 "kalabash_sievefilters:filter_move_down",
                 args=[setname, f["name"]]),
             "img": "fa fa-chevron-down",
             "title": _("Move this filter down")},
        ]
    return render_actions(actions)


@register.simple_tag
def display_errors(errors):
    if not len(errors):
        return ""
    t = template.Template("""
<ul class="errors">
  {% for error in errors %}
  <li>{{ error|safe }}</li>
  {% endfor %}
</ul>
""")
    return t.render(template.Context({
        "errors": errors
    }))


@register.simple_tag
def display_condition(form, cnt):
    target = form["cond_target_%d" % cnt]
    configure_field_classes(target)
    operator = form["cond_operator_%d" % cnt]
    configure_field_classes(operator)
    value = form["cond_value_%d" % cnt]
    configure_field_classes(value)
    t = template.Template("""
<div id="condition_{{ idx }}" class="item">
  <div class="col-lg-3 col-md-3 col-sm-3">{{ tfield }}</div>
  <div class="col-lg-3 col-md-3 col-sm-3">{{ofield}}</div>
  <div class="col-lg-4 col-md-4 col-sm-4">{{ vfield }}</div>
  {{ verrors }}
</div>
""")
    return t.render(template.Context({
        "idx": cnt,
        "tfield": target,
        "ofield": operator,
        "vfield": value,
        "verrors": display_errors(value.errors)
    }))


@register.simple_tag
def display_action(form, cnt):
    action = form["action_name_%d" % cnt]
    tpl = lib.find_action_template(action.initial)
    configure_field_classes(action)
    values = []
    verrors = []
    for pos, argtpl in enumerate(tpl.get("args", [])):
        fieldname = "action_arg_{}_{}".format(cnt, pos)
        if fieldname not in form.fields:
            continue
        field = form[fieldname]
        configure_field_classes(field)
        arg = {"field": field}
        if field.errors:
            verrors += field.errors
        if argtpl["type"] == "boolean":
            arg["checkbox"] = True
        values += [arg]
    t = template.Template("""
<div id="action_{{ idx }}" class="item">
<div class="col-sm-4">
  {{ afield }}
  </div>
  <div class="col-sm-6">
  {% for v in values %}{% if v.checkbox %}<div class="checkbox"><label>{{ v.field }} {{ v.field.label }}</label></div>{% else %}{{ v.field }}{% endif %}{% endfor %}
  {{ verrors }}
  </div>
</div>
""")
    return t.render(template.Context({
        "idx": cnt,
        "afield": action, "values": values,
        "verrors": display_errors(verrors)
    }))



# coding: utf-8
"""Declare and register the sievefilters extension."""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy

from kalabash.core.extensions import KbashExtension, exts_pool
from kalabash.parameters import tools as param_tools

from . import __version__
from . import forms


class SieveFilters(KbashExtension):
    name = "kalabash_sievefilters"
    label = gettext_lazy("Sieve filters")
    version = __version__
    description = gettext_lazy("Plugin to easily create server-side filters")
    url = "sfilters"
    topredirection_url = reverse_lazy("kalabash_sievefilters:index")

    def load(self):
        param_tools.registry.add(
            "global", forms.ParametersForm, gettext_lazy("Sieve filters"))
        param_tools.registry.add(
            "user", forms.UserSettings, gettext_lazy("Message filters"))


exts_pool.register_extension(SieveFilters)



# coding: utf-8

"""Mock objects."""


SAMPLE_SIEVE_SCRIPT = """require ["fileinto", "copy"];

# Filter: test1
if anyof (header :contains "To" "test1") {
    fileinto :copy "Test1";
}

# Filter: test2
if anyof (header :contains "To" "test2") {
    fileinto "Test2";
}
"""

SAMPLE_SIEVE_SCRIPT2 = """require ["fileinto"];

# Filter: Tést
if anyof (header :contains "Subject" "toto", not header :contains "To" "peé") {
    fileinto "Administratif.Contrats.CIC";
    stop;
}
"""


class ManagesieveClientMock(object):
    """Fake managesieve client."""

    def __init__(self, *args, **kwargs):
        self.scripts = {
            "main_script": SAMPLE_SIEVE_SCRIPT,
            "second_script": SAMPLE_SIEVE_SCRIPT,
            "third_script": SAMPLE_SIEVE_SCRIPT2,
        }

    def connect(self, username, password, **kwargs):
        return (username == "user@test.com" and password == "toto")

    def logout(self):
        return True

    def capability(self):
        return """
"IMPLEMENTATION" "Example1 ManageSieved v001"
"VERSION" "1.0"
"SASL" "PLAIN SCRAM-SHA-1 GSSAPI"
"SIEVE" "fileinto vacation"
"STARTTLS"
"""

    def havespace(self, name, size):
        return True

    def listscripts(self):
        return ("main_script", ["second_script"])

    def getscript(self, name):
        return self.scripts.get(name)

    def putscript(self, name, content):
        self.scripts[name] = content
        return True

    def setactive(self, name):
        return True

    def deletescript(self, name):
        return True


class IMAP4Mock(object):
    """Fake IMAP4 client."""

    def __init__(self, *args, **kwargs):
        self.untagged_responses = {}

    def _quote(self, data):
        return data

    def _simple_command(self, name, *args, **kwargs):
        if name == "CAPABILITY":
            self.untagged_responses["CAPABILITY"] = [b""]
        elif name == "LIST":
            self.untagged_responses["LIST"] = [b"() \".\" \"INBOX\""]
        return "OK", None

    def list(self):
        return "OK", [b"() \".\" \"INBOX\""]



"""Custom forms."""

from kalsievelib import commands
from kalsievelib.managesieve import SUPPORTED_AUTH_MECHS

from django import forms
from django.http import QueryDict
from django.utils.encoding import smart_str
from django.utils.translation import gettext as _, gettext_lazy

from kalabash.admin.templatetags.admin_tags import gender
from kalabash.lib import form_utils
from kalabash.parameters import forms as param_forms

from . import constants
from . import imaputils
from . import lib


class FiltersSetForm(forms.Form):
    name = forms.CharField(
        widget=forms.TextInput(attrs={"class": "form-control"})
    )
    active = forms.BooleanField(
        label=gender("Active", "m"), required=False,
        initial=False,
        help_text=gettext_lazy("Check to activate this filters set")
    )


class FilterForm(forms.Form):
    """A dynamic form to edit a filter."""

    def __init__(self, conditions, actions, request, *args, **kwargs):
        super(FilterForm, self).__init__(*args, **kwargs)

        self.field_widths = {
            "match_type": 8
        }

        self.fields["name"] = forms.CharField(label=_("Name"))
        self.fields["match_type"] = forms.ChoiceField(
            label=_("Match type"),
            choices=[("allof", _("All of the following")),
                     ("anyof", _("Any of the following")),
                     ("all", _("All messages"))],
            initial="anyof",
            widget=form_utils.HorizontalRadioSelect()
        )

        self.conds_cnt = 0
        for c in conditions:
            getattr(self, "_build_%s_field" % c[0])(c[1], c[2])
        self.actions_cnt = 0
        for a in actions:
            getattr(self, "_build_%s_field" % a[0])(request, *a[1:])

    def clean_name(self):
        """Check that name does not contain strange chars."""
        if "#" in self.cleaned_data["name"]:
            raise forms.ValidationError(_("Wrong filter name"))
        return self.cleaned_data["name"]

    def _build_header_field(self, name, op, value):
        """Add a new header field to form."""
        targets = []
        ops = []
        vfield = None
        for tpl in constants.CONDITION_TEMPLATES:
            targets += [(tpl["name"], tpl["label"]), ]
            if tpl["name"] != name:
                continue
            for opdef in tpl["operators"]:
                ops += [opdef[:2]]
                if op != opdef[0]:
                    continue
                if opdef[2] in ["string", "number"]:
                    vfield = forms.CharField(max_length=255, initial=value)

        self.fields["cond_target_%d" % self.conds_cnt] = \
            forms.ChoiceField(initial=name, choices=targets)
        self.fields["cond_operator_%d" % self.conds_cnt] = \
            forms.ChoiceField(initial=op, choices=ops)
        self.fields["cond_value_%d" % self.conds_cnt] = vfield
        self.conds_cnt += 1

    def _build_Subject_field(self, op, value):
        self._build_header_field("Subject", op, value)

    def _build_To_field(self, op, value):
        self._build_header_field("To", op, value)

    def _build_From_field(self, op, value):
        self._build_header_field("From", op, value)

    def _build_Cc_field(self, op, value):
        self._build_header_field("Cc", op, value)

    def _build_size_field(self, op, value):
        self._build_header_field("size", op, value)

    def _build_action_field(self, request, name, *values):
        """Add a new action field to form."""
        actions = []
        args = None
        for tpl in constants.ACTION_TEMPLATES:
            actions += [(tpl["name"], tpl["label"]), ]
            if name == tpl["name"]:
                args = tpl.get("args", [])
        self.fields["action_name_%d" % self.actions_cnt] = (
            forms.ChoiceField(initial=name, choices=actions))
        for cnt in range(len(args)):
            arg = args[cnt]
            value = values[cnt] if len(values) > cnt else None
            aname = "action_arg_%d_%d" % (self.actions_cnt, cnt)
            if arg["type"] == "string":
                self.fields[aname] = forms.CharField(
                    max_length=255,
                    initial=value)
            elif arg["type"] == "boolean":
                self.fields[aname] = forms.BooleanField(
                    label=arg["label"], initial=value, required=False)
            elif arg["type"] == "list":
                choices = getattr(self, arg["vloader"])(request)
                self.fields[aname] = forms.ChoiceField(
                    initial=value,
                    choices=choices)
        self.actions_cnt += 1

    def _build_redirect_field(self, request, *values):
        self._build_action_field(request, "redirect", *values)

    def _build_reject_field(self, request, *values):
        self._build_action_field(request, "reject", *values)

    def _build_fileinto_field(self, request, *values):
        self._build_action_field(request, "fileinto", *values)

    def _build_stop_field(self, request):
        self._build_action_field(request, "stop")

    def __build_folders_list(self, folders, user, imapc, parentmb=None):
        ret = []
        for fd in folders:
            value = fd["path"] if "path" in fd else fd["name"]
            if parentmb:
                ret += [
                    (value, fd["name"].replace(
                        "%s%s" % (parentmb, imapc.hdelimiter), ""))
                ]
            else:
                ret += [(value, fd["name"])]
            if "sub" in fd:
                submboxes = imapc.getmboxes(user, value)
                ret += self.__build_folders_list(submboxes, user, imapc, value)
        return ret

    def userfolders(self, request):
        mbc = imaputils.get_imapconnector(request)
        ret = mbc.getmboxes(request.user)

        folders = self.__build_folders_list(ret, request.user, mbc)
        return folders

    def tofilter(self):
        """Convert form values to filter values."""
        conditions = []
        actions = []
        for cpt in range(self.conds_cnt):
            conditions += [(
                self.cleaned_data["cond_target_%d" % cpt],
                ":" + self.cleaned_data["cond_operator_%d" % cpt],
                self.cleaned_data["cond_value_%d" % cpt])
            ]
        for cpt in range(self.actions_cnt):
            action = self.cleaned_data["action_name_%d" % cpt]
            tpl = lib.find_action_template(action)
            naction = [action]
            args = {}
            for pos, argtpl in enumerate(tpl.get("args", [])):
                fieldname = "action_arg_{}_{}".format(cpt, pos)
                if fieldname not in self.cleaned_data:
                    continue
                value = self.cleaned_data[fieldname]
                if argtpl["type"] == "boolean" and isinstance(value, bool):
                    if not value:
                        continue
                    value = argtpl["value"]
                args[argtpl["name"]] = value
            if "args_order" in tpl:
                # Corresponding command requires args to be in a special order
                for name in tpl["args_order"]:
                    if name in args:
                        naction.append(args[name])
            else:
                naction += [arg for arg in args.values()]
            actions += [naction]
        return (conditions, actions)


def build_filter_form_from_qdict(request):
    conditions = []
    actions = []
    qdict = QueryDict("", mutable=True)
    qdict["name"] = request.POST["name"]
    qdict["match_type"] = request.POST["match_type"]
    cpt = 0
    i = 0
    if qdict["match_type"] != "all":
        while True:
            if cpt == int(request.POST["conds"]):
                break
            if "cond_target_%d" % i in request.POST:
                qdict["cond_target_%d" % cpt] = (
                    request.POST["cond_target_%d" % i]
                )
                qdict["cond_operator_%d" % cpt] = (
                    request.POST["cond_operator_%d" % i]
                )
                qdict["cond_value_%d" % cpt] = (
                    request.POST["cond_value_%d" % i])
                condtarget = request.POST["cond_target_%d" % i]
                condop = request.POST["cond_operator_%d" % i]
                condvalue = request.POST["cond_value_%d" % i]
                conditions += [(condtarget, condop, condvalue)]
                cpt += 1
            i += 1
    cpt = 0
    i = 0
    while True:
        if cpt == int(request.POST["actions"]):
            break
        if "action_name_%d" % i in request.POST:
            qdict["action_name_%d" % cpt] = request.POST["action_name_%d" % i]
            action = request.POST["action_name_%d" % i]
            argcpt = 0
            args = []
            while True:
                try:
                    qdict["action_arg_%d_%d" % (cpt, argcpt)] = \
                        request.POST["action_arg_%d_%d" % (i, argcpt)]
                    args += [request.POST["action_arg_%d_%d" % (i, argcpt)]]
                except KeyError:
                    break
                argcpt += 1
            args = [action] + args
            actions += [args]
            cpt += 1
        i += 1

    return FilterForm(conditions, actions, request, qdict)


def build_filter_form_from_filter(request, name, fobj):
    match_type = fobj["test"].name
    conditions = []
    for t in fobj["test"]["tests"]:
        if isinstance(t, commands.TrueCommand):
            match_type = "all"
            conditions += [("Subject", "contains", "")]
            break
        elif isinstance(t, commands.SizeCommand):
            conditions += [("size", t["comparator"][1:], t["limit"])]
        else:
            operator_prefix = ""
            if isinstance(t, commands.NotCommand):
                t = t["test"]
                operator_prefix = "not"
            conditions += [(
                smart_str(t["header-names"]).strip('"'),
                "{}{}".format(
                    operator_prefix, smart_str(t["match-type"])[1:]),
                smart_str(t["key-list"]).strip('"'))
            ]
    actions = []
    for c in fobj.children:
        action = (c.name,)
        tpl = lib.find_action_template(c.name)
        for argtpl in tpl.get("args", []):
            if argtpl["name"] not in c:
                continue
            value = c[argtpl["name"]]
            if argtpl["type"] == "boolean":
                value = True if value else False
            else:
                value = value.strip('"')
            action += (value, )
        actions += [action]
    form = FilterForm(conditions, actions, request)
    form.fields["name"].initial = smart_str(name)
    form.fields["match_type"].initial = match_type
    return form


def supported_auth_mechs():
    values = [('AUTO', 'auto')]
    for m in SUPPORTED_AUTH_MECHS:
        values += [(m, m.lower())]
    return values


class ParametersForm(param_forms.AdminParametersForm):
    app = "kalabash_sievefilters"

    sep1 = form_utils.SeparatorField(
        label=gettext_lazy("ManageSieve settings"))

    server = forms.CharField(
        label=gettext_lazy("Server address"),
        initial="127.0.0.1",
        help_text=gettext_lazy("Address of your MANAGESIEVE server"),
        widget=forms.TextInput(attrs={"class": "form-control"})
    )

    port = forms.IntegerField(
        label=gettext_lazy("Server port"),
        initial=4190,
        help_text=gettext_lazy("Listening port of your MANAGESIEVE server"),
        widget=forms.TextInput(attrs={"class": "form-control"})
    )

    starttls = form_utils.YesNoField(
        label=gettext_lazy("Connect using STARTTLS"),
        initial=False,
        help_text=gettext_lazy("Use the STARTTLS extension")
    )

    authentication_mech = forms.ChoiceField(
        label=gettext_lazy("Authentication mechanism"),
        choices=supported_auth_mechs(),
        initial="auto",
        help_text=gettext_lazy("Prefered authentication mechanism"),
        widget=forms.Select(attrs={"class": "form-control"})
    )

    sep2 = form_utils.SeparatorField(label=gettext_lazy("IMAP settings"))

    imap_server = forms.CharField(
        label=gettext_lazy("Server address"),
        initial="127.0.0.1",
        help_text=gettext_lazy("Address of your IMAP server")
    )

    imap_secured = form_utils.YesNoField(
        label=gettext_lazy("Use a secured connection"),
        initial=False,
        help_text=gettext_lazy(
            "Use a secured connection to access IMAP server")
    )

    imap_port = forms.IntegerField(
        label=gettext_lazy("Server port"),
        initial=143,
        help_text=gettext_lazy("Listening port of your IMAP server")
    )


class UserSettings(param_forms.UserParametersForm):
    app = "kalabash_sievefilters"

    sep1 = form_utils.SeparatorField(label=gettext_lazy("General"))

    editor_mode = forms.ChoiceField(
        initial="gui",
        label=gettext_lazy("Editor mode"),
        choices=[("raw", "raw"), ("gui", "simplified")],
        help_text=gettext_lazy(
            "Select the mode you want the editor to work in"),
        widget=form_utils.HorizontalRadioSelect()
    )

    sep2 = form_utils.SeparatorField(label=gettext_lazy("Mailboxes"))

    trash_folder = forms.CharField(
        initial="Trash",
        label=gettext_lazy("Trash folder"),
        help_text=gettext_lazy("Folder where deleted messages go")
    )

    sent_folder = forms.CharField(
        initial="Sent",
        label=gettext_lazy("Sent folder"),
        help_text=gettext_lazy("Folder where copies of sent messages go")
    )

    drafts_folder = forms.CharField(
        initial="Drafts",
        label=gettext_lazy("Drafts folder"),
        help_text=gettext_lazy("Folder where drafts go")
    )

    @staticmethod
    def has_access(**kwargs):
        return hasattr(kwargs.get("user"), "mailbox")



# -*- coding: utf-8 -*-

"""The sieve filters editor of Kalabash."""

from __future__ import unicode_literals

from pkg_resources import get_distribution, DistributionNotFound


try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

default_app_config = "kalabash_sievefilters.apps.SieveFiltersConfig"



"""Sieve urls."""

from django.urls import path

from . import views

app_name = 'kalabash_sievefilters'

urlpatterns = [
    path('', views.index, name="index"),
    path('savefs/<str:name>/', views.savefs, name="fs_save"),
    path('newfs/', views.new_filters_set, name="fs_add"),
    path('removefs/<str:name>/', views.remove_filters_set,
         name="fs_delete"),
    path('activatefs/<str:name>/', views.activate_filters_set,
         name="fs_activate"),
    path('downloadfs/<str:name>/', views.download_filters_set,
         name="fs_download"),
    path('templates/<str:ftype>/', views.get_templates,
         name="templates_get"),
    path('<str:setname>/newfilter/', views.newfilter,
         name="filter_add"),
    path('<str:setname>/editfilter/<path:fname>/', views.editfilter,
         name="filter_change"),
    path('<str:setname>/removefilter/<path:fname>/', views.removefilter,
         name="filter_delete"),
    path('<str:setname>/togglestate/<path:fname>/',
         views.toggle_filter_state, name="filter_toggle_state"),
    path('<str:setname>/moveup/<path:fname>/',
         views.move_filter_up, name="filter_move_up"),
    path('<str:setname>/movedown/<path:fname>/',
         views.move_filter_down, name="filter_move_down"),
    path('<str:name>/', views.getfs, name="fs_get"),
]



"""
Copy of the original rfc6266 package, just the generation part.

Orginal code is available at https://github.com/g2p/rfc6266.
"""

from urllib.parse import quote


# RFC 2616
separator_chars = "()<>@,;:\\\"/[]?={} \t"
# RFC 5987
attr_chars_nonalnum = '!#$&+-.^_`|~'


def percent_encode(string, safe, encoding):
    return quote(string, safe, encoding, errors='strict')


def is_token_char(ch):
    # Must be ascii, and neither a control char nor a separator char
    asciicode = ord(ch)
    # < 128 means ascii, exclude control chars at 0-31 and 127,
    # exclude separator characters.
    return 31 < asciicode < 127 and ch not in separator_chars


def is_token(candidate):
    return all(is_token_char(ch) for ch in candidate)


def is_ascii(text):
    return all(ord(ch) < 128 for ch in text)


def fits_inside_codec(text, codec):
    try:
        text.encode(codec)
    except UnicodeEncodeError:
        return False
    else:
        return True


def is_lws_safe(text):
    return normalize_ws(text) == text


def normalize_ws(text):
    return ' '.join(text.split())


def qd_quote(text):
    return text.replace('\\', '\\\\').replace('"', '\\"')


def build_header(
    filename, disposition='attachment', filename_compat=None
):
    """Generate a Content-Disposition header for a given filename.
    For legacy clients that don't understand the filename* parameter,
    a filename_compat value may be given.
    It should either be ascii-only (recommended) or iso-8859-1 only.
    In the later case it should be a character string
    (unicode in Python 2).
    Options for generating filename_compat (only useful for legacy clients):
    - ignore (will only send filename*);
    - strip accents using unicode's decomposing normalisations,
    which can be done from unicode data (stdlib), and keep only ascii;
    - use the ascii transliteration tables from Unidecode (PyPI);
    - use iso-8859-1
    Ignore is the safest, and can be used to trigger a fallback
    to the document location (which can be percent-encoded utf-8
    if you control the URLs).
    See https://tools.ietf.org/html/rfc6266#appendix-D
    """

    # While this method exists, it could also sanitize the filename
    # by rejecting slashes or other weirdness that might upset a receiver.

    if disposition != 'attachment':
        assert is_token(disposition)

    rv = disposition

    if is_token(filename):
        rv += '; filename=%s' % (filename, )
        return rv
    elif is_ascii(filename) and is_lws_safe(filename):
        qd_filename = qd_quote(filename)
        rv += '; filename="%s"' % (qd_filename, )
        if qd_filename == filename:
            # RFC 6266 claims some implementations are iffy on qdtext's
            # backslash-escaping, we'll include filename* in that case.
            return rv
    elif filename_compat:
        if is_token(filename_compat):
            rv += '; filename=%s' % (filename_compat, )
        else:
            assert is_lws_safe(filename_compat)
            rv += '; filename="%s"' % (qd_quote(filename_compat), )

    # alnum are already considered always-safe, but the rest isn't.
    # Python encodes ~ when it shouldn't, for example.
    rv += "; filename*=utf-8''%s" % (percent_encode(
        filename, safe=attr_chars_nonalnum, encoding='utf-8'), )

    # This will only encode filename_compat, if it used non-ascii iso-8859-1.
    return rv.encode('iso-8859-1')



# coding: utf-8

"""Sievefilters tests."""

from unittest import mock

from django.urls import reverse

from kalabash.admin import factories as admin_factories
from kalabash.core import models as core_models
from kalabash.lib.tests import KbashTestCase

from . import mocks


class SieveFiltersTestCase(KbashTestCase):
    """Check sieve filters."""

    @classmethod
    def setUpTestData(cls):
        """Create some users."""
        super(SieveFiltersTestCase, cls).setUpTestData()
        admin_factories.populate_database()
        cls.user = core_models.User.objects.get(username="user@test.com")

    def setUp(self):
        """Connect with a simpler user."""
        patcher = mock.patch("kalsievelib.managesieve.Client")
        self.mock_client = patcher.start()
        self.mock_client.return_value = mocks.ManagesieveClientMock()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("imaplib.IMAP4")
        self.mock_imap4 = patcher.start()
        self.mock_imap4.return_value = mocks.IMAP4Mock()
        self.addCleanup(patcher.stop)

        url = reverse("core:login")
        data = {
            "username": self.user.username, "password": "toto"
        }
        self.client.post(url, data)

    def test_index(self):
        """Test index view."""
        response = self.client.get(reverse("kalabash_sievefilters:index"))
        self.assertContains(response, "main_script (active)")
        response = self.client.get(reverse("core:logout"))
        self.assertEqual(response.status_code, 302)

    def test_getfs(self):
        """Test getfs view."""
        response = self.ajax_get(
            reverse("kalabash_sievefilters:fs_get", args=["main_script"]))
        self.assertIn(
            "/sfilters/main_script/editfilter/test1/", response["content"])

    def test_toggle_filter_state(self):
        """Test toggle_filter_state view."""
        url = reverse("kalabash_sievefilters:filter_toggle_state",
                      args=["main_script", "test1"])
        response = self.ajax_get(url)
        self.assertEqual(response["color"], "red")
        response = self.ajax_get(url)
        self.assertEqual(response["color"], "green")

    def test_move_filter_up(self):
        """Test move_filter_up view."""
        url = reverse("kalabash_sievefilters:filter_move_up",
                      args=["main_script", "test2"])
        response = self.ajax_get(url)
        pos1 = response["content"].find(
            "/sfilters/main_script/editfilter/test2/")
        pos2 = response["content"].find(
            "/sfilters/main_script/editfilter/test1/")
        self.assertTrue(pos1 < pos2)

    def test_move_filter_down(self):
        """Test move_filter_down view."""
        url = reverse("kalabash_sievefilters:filter_move_down",
                      args=["main_script", "test1"])
        response = self.ajax_get(url)
        pos1 = response["content"].find(
            "/sfilters/main_script/editfilter/test2/")
        pos2 = response["content"].find(
            "/sfilters/main_script/editfilter/test1/")
        self.assertTrue(pos1 < pos2)

    def test_new_filter(self):
        """Test new_filter view."""
        url = reverse("kalabash_sievefilters:filter_add", args=["main_script"])
        response = self.client.get(url)
        self.assertContains(response, "New filter")

        data = {
            "name": "tést filter",
            "match_type": "anyof",
            "conds": 1,
            "cond_target_0": "Subject",
            "cond_operator_0": "contains",
            "cond_value_0": "Test",
            "actions": 1,
            "action_name_0": "fileinto",
            "action_arg_0_0": "Trash",
        }
        response = self.ajax_post(url, data)
        self.assertEqual(response, "Filter created")

    def test_edit_filter(self):
        """Test edit_filter view."""
        url = reverse(
            "kalabash_sievefilters:filter_change",
            args=["main_script", "test1"])
        response = self.client.get(url)
        self.assertContains(response, "Edit filter")

        data = {
            "oldname": "test1",
            "name": "test1",
            "match_type": "all",
            "conds": 1,
            "cond_target_0": "To",
            "cond_operator_0": "contains",
            "cond_value_0": "test1",
            "actions": 1,
            "action_name_0": "fileinto",
            "action_arg_0_0": "Trash",
            "action_arg_0_1": True
        }
        response = self.ajax_post(url, data)
        self.assertEqual(response, "Filter modified")

    def test_edit_filter_with_not(self):
        """Try to edit a filter containing a not operator."""
        url = reverse(
            "kalabash_sievefilters:filter_change",
            args=["third_script", "Tést"])
        response = self.client.get(url)
        self.assertContains(response, "Edit filter")

    def test_remove_filter(self):
        """Test removefilter view."""
        url = reverse(
            "kalabash_sievefilters:filter_delete",
            args=["second_script", "test1"])
        response = self.ajax_get(url)
        self.assertEqual(response, "Filter removed")

    def test_download_filters_set(self):
        """Test download_filters_set view."""
        url = reverse("kalabash_sievefilters:fs_download", args=["main_script"])
        response = self.client.get(url)
        self.assertEqual(response.content.decode(), mocks.SAMPLE_SIEVE_SCRIPT)

    def test_activate_filters_set(self):
        """Test activate_filters_set view."""
        url = reverse(
            "kalabash_sievefilters:fs_activate", args=["second_script"])
        response = self.ajax_get(url)
        self.assertEqual(response["respmsg"], "Filters set activated")

    def test_remove_filters_set(self):
        """Test remove_filters_set view."""
        url = reverse(
            "kalabash_sievefilters:fs_delete", args=["second_script"])
        response = self.ajax_get(url)
        self.assertEqual(response["respmsg"], "Filters set deleted")

    def test_new_filters_set(self):
        """Test new_filters_set view."""
        url = reverse("kalabash_sievefilters:fs_add")
        response = self.client.get(url)
        self.assertContains(response, "Create a new filters set")

        data = {"name": "new_script"}
        response = self.ajax_post(url, data)
        self.assertEqual(response["respmsg"], "Filters set created")

    def test_savefs(self):
        """Test savefs view."""
        url = reverse("kalabash_sievefilters:fs_save", args=["new_script"])
        data = {"scriptcontent": mocks.SAMPLE_SIEVE_SCRIPT}
        response = self.ajax_post(url, data)
        self.assertEqual(response["respmsg"], "Filters set saved")



# coding: utf-8

"""Custom views."""

from kalsievelib.commands import BadArgument, BadValue
from kalsievelib.managesieve import Error

from django.contrib.auth.decorators import login_required
from django.urls import reverse
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from django.utils.encoding import smart_bytes, smart_str
from django.utils.translation import gettext as _

from kalabash.admin.lib import needs_mailbox
from kalabash.lib.connections import ConnectionError
from kalabash.lib.exceptions import BadRequest
from kalabash.lib.web_utils import (
    _render_error, ajax_response, render_to_json_response
)

from . import constants
from .forms import (
    FilterForm, build_filter_form_from_qdict, build_filter_form_from_filter,
    FiltersSetForm
)
from .imaputils import get_imapconnector
from .lib import SieveClient, SieveClientError
from .rfc6266 import build_header
from .templatetags.sfilters_tags import fset_menu


@login_required
@needs_mailbox()
def index(request, tplname="kalabash_sievefilters/index.html"):
    try:
        sc = SieveClient(user=request.user.username,
                         password=request.session["password"])
    except ConnectionError as e:
        return _render_error(request, user_context={"error": e})

    try:
        active_script, scripts = sc.listscripts()
    except Error as e:
        return _render_error(request, user_context={"error": e})

    if active_script is None:
        active_script = ""
        default_script = "%s/" % scripts[0] if len(scripts) else ""
    else:
        default_script = "%s/" % active_script
    return render(request, tplname, {
        "selection": "user",
        "active_script": active_script,
        "default_script": default_script,
        "scripts": sorted(scripts),
        "hdelimiter": get_imapconnector(request).hdelimiter
    })


@login_required
@needs_mailbox()
def get_templates(request, ftype):
    if ftype == "condition":
        data = constants.CONDITION_TEMPLATES
    else:
        data = constants.ACTION_TEMPLATES
    return JsonResponse(data, safe=False)


@login_required
@needs_mailbox()
def getfs(request, name):
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    editormode = request.user.parameters.get_value("editor_mode")
    error = None
    try:
        content = sc.getscript(name, format=editormode)
    except SieveClientError as e:
        error = str(e)
    else:
        if content is None:
            error = _("Failed to retrieve filters set")

    if error is not None:
        return ajax_response(request, "ko", respmsg=error)

    if editormode == "raw":
        htmlcontent = render_to_string(
            "kalabash_sievefilters/rawfilter.html", {
                "name": name, "scriptcontent": content
            })
    else:
        htmlcontent = render_to_string(
            "kalabash_sievefilters/guieditor.html", {"fs": content})

    menu = (
        "<ul id='fsetmenu' class='nav nav-sidebar'>"
        "<li class='nav-header'>{}</li>{}</ul>".format(
            _("Actions"), fset_menu(editormode, name)
        )
    )
    resp = {'menu': menu, 'content': htmlcontent}
    return render_to_json_response(resp)


def build_filter_ctx(ctx, form):
    ctx["form"] = form
    ctx["conds_nb"] = range(form.conds_cnt)
    ctx["actions_nb"] = range(form.actions_cnt)
    return ctx


def submitfilter(
        request, setname, okmsg, tplname, tplctx, update=False, sc=None):
    form = build_filter_form_from_qdict(request)
    if form.is_valid():
        if sc is None:
            sc = SieveClient(user=request.user.username,
                             password=request.session["password"])
        fset = sc.getscript(setname, format="fset")
        conditions, actions = form.tofilter()
        match_type = form.cleaned_data["match_type"]
        if match_type == "all":
            match_type = "anyof"
            conditions = [("true",)]
        fltname = form.cleaned_data["name"].encode("utf-8")
        try:
            if not update:
                fset.addfilter(fltname, conditions, actions,
                               match_type)
            else:
                oldname = request.POST["oldname"].encode("utf-8")
                fset.updatefilter(
                    oldname, fltname, conditions, actions, match_type
                )
        except (BadArgument, BadValue) as inst:
            raise BadRequest(str(inst))
        sc.pushscript(fset.name, "{}".format(fset))
        return render_to_json_response(okmsg)

    return render_to_json_response({'form_errors': form.errors}, status=400)


@login_required
@needs_mailbox()
def newfilter(request, setname, tplname="kalabash_sievefilters/filter.html"):
    ctx = {
        "title": _("New filter"),
        "formid": "filterform",
        "action": reverse("kalabash_sievefilters:filter_add", args=[setname]),
        "action_label": _("Create"),
        "action_classes": "submit"
    }
    if request.method == "POST":
        return submitfilter(
            request, setname, _("Filter created"), tplname, ctx
        )

    conds = [("Subject", "contains", "")]
    actions = [("fileinto", "")]
    form = FilterForm(conds, actions, request)
    ctx = build_filter_ctx(ctx, form)
    return render(request, tplname, ctx)


@login_required
@needs_mailbox()
def editfilter(request, setname, fname, tplname="kalabash_sievefilters/filter.html"):
    ctx = {
        "title": _("Edit filter"),
        "formid": "filterform",
        "action": reverse("kalabash_sievefilters:filter_change",
                          args=[setname, fname]),
        "action_label": _("Update"),
        "action_classes": "submit"
    }
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    if request.method == "POST":
        return submitfilter(
            request, setname, _("Filter modified"), tplname, ctx,
            update=True, sc=sc
        )

    fset = sc.getscript(setname, format="fset")
    fname = smart_bytes(fname)
    f = fset.getfilter(fname)
    form = build_filter_form_from_filter(request, fname, f)
    ctx = build_filter_ctx(ctx, form)
    ctx["oldname"] = fname
    ctx["hidestyle"] = "none" \
        if form.fields["match_type"].initial == "all" else "block"
    return render(request, tplname, ctx)


@login_required
@needs_mailbox()
def removefilter(request, setname, fname):
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    fset = sc.getscript(setname, format="fset")
    if fset.removefilter(fname.encode("utf-8")):
        sc.pushscript(fset.name, "{}".format(fset))
        return render_to_json_response(_("Filter removed"))
    return render_to_json_response(_("Failed to remove filter"), status=500)


@login_required
@needs_mailbox()
def savefs(request, name):
    if "scriptcontent" not in request.POST:
        return
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    try:
        sc.pushscript(name, request.POST["scriptcontent"])
    except SieveClientError as e:
        error = str(e)
        return ajax_response(request, "ko", respmsg=error)
    return ajax_response(request, respmsg=_("Filters set saved"))


@login_required
@needs_mailbox()
def new_filters_set(request, tplname="common/generic_modal_form.html"):
    if request.method == "POST":
        form = FiltersSetForm(request.POST)
        if form.is_valid():
            sc = SieveClient(user=request.user.username,
                             password=request.session["password"])
            sc.pushscript(form.cleaned_data["name"], "# Empty script",
                          form.cleaned_data["active"])
            return render_to_json_response({
                "url": form.cleaned_data["name"],
                "active": form.cleaned_data["active"],
                "respmsg": _("Filters set created")
            })
        return render_to_json_response(
            {'form_errors': form.errors}, status=400)

    ctx = {"title": _("Create a new filters set"),
           "formid": "newfiltersset",
           "action_label": _("Create"),
           "action_classes": "submit",
           "action": reverse("kalabash_sievefilters:fs_add"),
           "withmenu": False,
           "withunseen": False,
           "form": FiltersSetForm()}
    return render(request, tplname, ctx)


@login_required
@needs_mailbox()
def remove_filters_set(request, name):
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    sc.deletescript(name)
    acs, scripts = sc.listscripts()
    ctx = {
        "respmsg": _("Filters set deleted"),
        "newfs": acs,
    }
    return render_to_json_response(ctx)


@login_required
@needs_mailbox()
def activate_filters_set(request, name):
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    try:
        sc.activatescript(name)
    except SieveClientError as e:
        return ajax_response(request, "ko", respmsg=str(e))
    return ajax_response(request, respmsg=_("Filters set activated"))


@login_required
@needs_mailbox()
def download_filters_set(request, name):
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    try:
        script = sc.getscript(name)
    except SieveClientError as e:
        return ajax_response(request, "ko", respmsg=str(e))

    resp = HttpResponse(script)
    resp["Content-Type"] = "text/plain; charset=utf-8"
    resp["Content-Length"] = len(script)
    resp["Content-Disposition"] = build_header('%s.txt' % name)
    return resp


@login_required
@needs_mailbox()
def toggle_filter_state(request, setname, fname):
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    fname = smart_str(fname)
    fset = sc.getscript(setname, format="fset")
    if fset.is_filter_disabled(fname):
        ret = fset.enablefilter(fname)
        newstate = _("yes")
        color = "green"
    else:
        ret = fset.disablefilter(fname)
        newstate = _("no")
        color = "red"
    if not ret:
        pass
    sc.pushscript(setname, "{}".format(fset))
    return render_to_json_response({
        "label": newstate,
        "color": color
    })


def move_filter(request, setname, fname, direction):
    sc = SieveClient(user=request.user.username,
                     password=request.session["password"])
    fset = sc.getscript(setname, format="fset")
    fset.movefilter(fname.encode("utf-8"), direction)
    sc.pushscript(setname, str(fset))
    return ajax_response(
        request, template="kalabash_sievefilters/guieditor.html", fs=fset
    )


@login_required
@needs_mailbox()
def move_filter_up(request, setname, fname):
    return move_filter(request, setname, fname, "up")


@login_required
@needs_mailbox()
def move_filter_down(request, setname, fname):
    return move_filter(request, setname, fname, "down")



# coding: utf-8

"""Internal tools."""

import ssl

from kalsievelib.factory import FiltersSet
from kalsievelib import managesieve
from kalsievelib.parser import Parser
import six

from django.utils.translation import gettext as _

from kalabash.lib.connections import ConnectionsManager, ConnectionError
from kalabash.lib.exceptions import KalabashException
from kalabash.parameters import tools as param_tools

from . import constants


class SieveClientError(KalabashException):
    http_code = 424


@six.add_metaclass(ConnectionsManager)
class SieveClient(object):
    """Sieve client."""

    def __init__(self, user=None, password=None):
        try:
            ret, msg = self.login(user, password)
        except managesieve.Error as e:
            raise ConnectionError(str(e))
        if not ret:
            raise ConnectionError(msg)

    def login(self, user, password):
        conf = dict(param_tools.get_global_parameters("kalabash_sievefilters"))
        self.msc = managesieve.Client(
            conf["server"], conf["port"], debug=False)
        authmech = conf["authentication_mech"]
        if authmech == "AUTO":
            authmech = None
        try:
            ret = self.msc.connect(user, password,
                                   starttls=conf["starttls"], authmech=authmech)
        except managesieve.Error:
            ret = False
        if not ret:
            return False, _(
                "Connection to MANAGESIEVE server failed, check your "
                "configuration"
            )
        return True, None

    def logout(self):
        self.msc.logout()
        self.msc = None

    def refresh(self, user, password):
        if self.msc is not None:
            try:
                self.msc.capability()
            except managesieve.Error as e:
                pass
            else:
                return
        try:
            ret, msg = self.login(user, password)
        except (managesieve.Error, ssl.SSLError) as e:
            raise ConnectionError(e)
        if not ret:
            raise ConnectionError(msg)

    def listscripts(self):
        return self.msc.listscripts()

    def getscript(self, name, format="raw"):
        content = self.msc.getscript(name)
        if content is None:
            raise SieveClientError(self.msc.errmsg.decode())
        if format == "raw":
            return content
        p = Parser()
        if not p.parse(content):
            print("Parse error????")
            return None
        fs = FiltersSet(name)
        fs.from_parser_result(p)
        return fs

    def pushscript(self, name, content, active=False):
        if not self.msc.havespace(name, len(content)):
            error = "%s (%s)" % (
                _("Not enough space on server"), self.msc.errmsg)
            raise SieveClientError(error)
        if not self.msc.putscript(name, content):
            raise SieveClientError(self.msc.errmsg.decode())
        if active and not self.msc.setactive(name):
            raise SieveClientError(self.msc.errmsg)

    def deletescript(self, name):
        if not self.msc.deletescript(name):
            raise SieveClientError(self.msc.errmsg.decode())

    def activatescript(self, name):
        if not self.msc.setactive(name):
            raise SieveClientError(self.msc.errmsg.decode())


def find_action_template(action):
    """Find template corresponding to action."""
    for tpl in constants.ACTION_TEMPLATES:
        if tpl["name"] == action:
            return tpl
    raise RuntimeError("action {} not defined".format(action))
